#ifndef lint
static char SCCSid[] = "@(#) ./comm/ialog.c 12/01/93";
#endif

/* This file contains routines to initialize the log files */
/*
  In previous versions, this was enabled only for the profiling version.
  However, that is rather restrictive.  Further, compile-time control,
  while useful in keeping the size of the image down (and removing 
  extra costs from the computation), does not permit easy debugging
  by runtime control (such as a command line parameter to enable 
  communication logging).

  In addition, it is useful to be able to control whether Chameleon operations
  are considered atomic or not.  For example, a gscatter or the setup code.

  Enhancements to consider:
  Pass the file and line to the event log routine so that the event file can
  contain the file and line number that generated the event.  The AIMS (Ames
  InstruMentation System does this, allowing you to click on an event and
  get the source code displayed).
 */

#include <string.h>
#include "tools.h"

#if !defined(LOGCOMMDISABLE) && !defined(LOGCOMMEVENTS) && !defined(LOGCOMMALL)
#define LOGCOMMEVENTS
#endif

#include "comm/comm.h"
/* global.h is needed to get a valid system message type */
#include "comm/global/global.h"
#include <stdio.h>

/* Define the logging level.  For now, all logging is disabled unless
   specifically enabled.  */
int        OnlyAtomic = 1, inAtomicOp = 0;
static int MergeLogFiles = 1;
int        cloglevel = 0;
static int initlog   = 0;
static int didinit   = 0;
FILE *_tracefile     = stdout;
/* Min tag and max tag are used to speed the test of the valid tag ranges.
   Note that we'd like to test maxtag against the values used for the 
   collective operations to insure that the user does not interfere with
   them; to do so, we need to allow the collective routines to bypass this
   test (or rather, use a different value) */
static int mintag = -1, maxtag = 0x7fffffff;

/* By making the message buffer a static, we can print the last entry out
   on an exception if the trace option is enabled.  The same buffer is used
   for error messages */
static char buf[256];

/* This routine allows you to log operations within a global operation with 
   flag = 0.  Set this with -traceall */
int PIiSetLogAtomic( flag )
int flag;
{
int old = OnlyAtomic;

OnlyAtomic = flag;
return old;
}

/*
    Routine to modify the tag boungs
 */
PIiSetTagLimits( min, max )
int min, max;
{
mintag = min;
maxtag = max;
}

/*@
  PISetMergeEventFiles - Control whether event files are merged

  Input Parameter:
. flag - 
@*/
void PISetMergeEventFiles( flag )
int flag;
{
MergeLogFiles = flag;
}

#define CLOG_EVENT   0x1
#define CLOG_TRACE   0x2
#define CLOG_SUMMARY 0x4
#define CLOG_LASTOP  0x8
#define CLOG_IDLE    0x10

/*@
   PISetLogging - Set the logging level for communications

   Input Parameter:
.  level - level to use:
$             0 - no logging
$             1 - event logs (to a file for postmortum use)
$             2 - print out for each send and recv
$             4 - keep a summary of times and counts
$             8 - print out LAST send or recv on error exit (interrupt)
$            16 - collect recv idle time (may require a busy loop)
$   Or the bits together to get the combinations (ie, 3 is print + log).
@*/
void PISetLogging( level )
int level;
{
cloglevel = level;
if (cloglevel & CLOG_EVENT) LOGENABLE;
else                        LOGDISABLE;
}
/*@
   PISetLoggingBit - Set/Clear the logging level for communications, for a 
                     single option

   Input Parameter:
.  bit - option to set
$             0 - no logging
$             1 - event logs (to a file for postmortum use)
$             2 - print out for each send and recv
$             4 - keep a summary of times and counts
$            32 - clear summary information (flag ignored)
$   Or the bits together to get the combinations (ie, 3 is print + log).
.  flag - whether to set or clear the option.  1 = set, 0 = clear
@*/
void PISetLoggingBit( bit, flag )
int bit, flag;
{
if (bit & 32) {
    PISummaryClear();
    return;
    }
if (flag) {
    cloglevel |= bit;
    if (bit == CLOG_EVENT) { LOGENABLE; initlog = 1; }
    }
else {
    cloglevel = cloglevel & (~bit);
    if (bit == CLOG_EVENT) LOGDISABLE; /* ? initlog = 0; ? */
    }
}

/* 
   In the past, this routine was set up to either issue the logging events
   or not, defined at COMPILE time.  This made it impossible to select this
   at RUNTIME.  
 */
PIinitlog()
{
#if defined(LOGCOMMEVENTS) && !defined(LOGCOMMDISABLE)
if (initlog && !didinit) {
    didinit = 1;
    if (PImytid == 0) {
	/* printf( "I am the master logger\n" ); */
	BLOG_MASTER(BLOG_TRUNCATE);
	LOGDEFINE(LOG_BSENDS,"Start bsend");
	LOGDEFINE(LOG_BSENDE,"End   bsend");
	LOGDEFINE(LOG_NSENDS,"Start nsend");
	LOGDEFINE(LOG_NSENDE,"End   nsend");
	LOGDEFINE(LOG_WSENDS,"Start wsend");
	LOGDEFINE(LOG_WSENDE,"End   wsend");
	LOGDEFINE(LOG_BRECVS,"Start brecv");
	LOGDEFINE(LOG_BRECVE,"End   brecv");
	LOGDEFINE(LOG_NRECVS,"Start nrecv");
	LOGDEFINE(LOG_NRECVE,"End   nrecv");
	LOGDEFINE(LOG_WRECVS,"Start wrecv");
	LOGDEFINE(LOG_WRECVE,"End   wrecv");
	LOGDEFINE(LOG_IRECVS,"Start recv wait");
	LOGDEFINE(LOG_IRECVE,"End   recv wait");

	LOGDEFINE(LOG_GSUMS,"Start GSum");
	LOGDEFINE(LOG_GSUME,"End GSum");
	LOGDEFINE(LOG_GMAXS,"Start GMax");
	LOGDEFINE(LOG_GMAXE,"End GMax");
	LOGDEFINE(LOG_GMINS,"Start GMin");
	LOGDEFINE(LOG_GMINE,"End GMin");
	LOGDEFINE(LOG_GCOLS,"Start GCollection");
	LOGDEFINE(LOG_GCOLE,"End GCollection");
	LOGDEFINE(LOG_GSCATS,"Start GScatter");
	LOGDEFINE(LOG_GSCATE,"End GScatter");
	LOGDEFINE(LOG_GSYNCS,"Start GSync");
	LOGDEFINE(LOG_GSYNCE,"End GSync");
	LOGSTATE(LOG_BRECVS,LOG_BRECVE,"seagreen1","Brecv");
	LOGSTATE(LOG_WRECVS,LOG_WRECVE,"seagreen2","Wrecv");
	LOGSTATE(LOG_NRECVS,LOG_NRECVE,"seagreen3","Nrecv");
	LOGSTATE(LOG_BSENDS,LOG_BSENDE,"blue","Bsend");
	LOGSTATE(LOG_NSENDS,LOG_NSENDE,"lightblue","Nsend");
	LOGSTATE(LOG_WSENDS,LOG_WSENDE,"steelblue","Wsend");
	LOGSTATE(LOG_GSYNCS,LOG_GSYNCE,"yellow","Sync");
	LOGSTATE(LOG_GSUMS,LOG_GSUME,"purple","GlobalSum");
	LOGSTATE(LOG_GCOLS,LOG_GCOLE,"magenta","GlobalCol");
	LOGSTATE(LOG_GMINS,LOG_GMINE,"firebrick","GlobalMin");
	LOGSTATE(LOG_GMAXS,LOG_GMAXE,"Olivedrab","GlobalMax");
	LOGSTATE(LOG_GSCATS,LOG_GSCATE,"cyan","Scatter");
	LOGSTATE(LOG_GTOKS,LOG_GTOKE,"white","Gtoken");
	LOGSTATE(LOG_IRECVS,LOG_IRECVE,"red","RecvWait");
	}
    else {
	/* printf( "Calling alog_setup with %d\n", PImytid ); */
	BLOG_SETUP(BLOG_TRUNCATE);
	}
    /* Add the information needed to adjust the clocks */
    BLOG_FIX_CLOCKS();
    }
#endif
}

PIendlog()
{
#if !defined(LOGCOMMDISABLE)
if (initlog && didinit) {
    /* Add the sync event at the end of the run */
    PIgsync( PSAllProcs );
    LOGEVENT(EVENT_SYNC);
    
    if (MergeLogFiles) {
	xx_BLOG_adjusttimes();
	/* Merge replaces xx_BLOG_dump() */
	BLOGMergeSetup( BLOGIsPicl() ? "bl.trf" : "bl" );
	BLOGParallelMerge( (void (*)())0 );
	}
    else {
	BLOG_OUTPUT;
	}
    didinit = 0;
    }
/* This might be a nice place to generate some summary data */
if (cloglevel & CLOG_SUMMARY) {
    int i;
    /* Turn off logging of the PIgtoken */
    cloglevel = 0;
    for (i=0; i<=PInumtids; i++) {
	if (PIgtoken(PSAllProcs,i)) {
	    fprintf( stdout, "Summary for node %d:\n", PImytid );
	    PILogSummary( stdout );
	    }
	}
    /* Now, generate a COMBINED SUMMARY */
    if (PImytid == 0) 
	fprintf( stdout, "Combined summary for %d tasks:\n", PInumtids );
    PILogCombinedSummary( stdout );
    }
#endif
}

/* This routine outputs the state info line */
void PIiOutputState( fp, state, name, do_rate )
FILE    *fp;
SYstate *state;
char    *name;
int     do_rate;
{
double r;
if (state->count > 0) {
    fprintf( fp, "%20s%10d %10d %13.4le", 
	     name, state->count, state->data, state->time );
    if (do_rate) {
	if (state->time > 0) r = state->data / state->time;
	else 		     r = 0.0;
	fprintf( fp, " %12.4le\n", r );
	}
    else {
	fprintf( fp, "\n" );
	}
    }
}	    

/* */
typedef struct {
    double maxtime, mintime;
    int    maxtask, mintask, maxcount, mincount, maxdata, mindata;
    double maxitime;
    int    maxline, maxitask;
    char   maxfile[SYSTATEFILELEN];
    } PIiStateSum;
/* This is an example of a routine that computes a fairly complicated
   combination function.  This finds (a) the maximum time and associate
   process, (b) the minimum time and the associated processor, and (c) the
   file, line, and processor for the single largest contribution.
 */
void PIiStateSumOp( a, b, n )
int         n;
PIiStateSum *a, *b;
{
if (a->maxtime < b->maxtime) {
    a->maxtime  = b->maxtime;
    a->maxtask  = b->maxtask;
    a->maxcount = b->maxcount;
    a->maxdata  = b->maxdata;
    }
if (a->maxitime < b->maxitime) {
    a->maxitime = b->maxitime;
    a->maxline  = b->maxline;
    strcpy(a->maxfile,b->maxfile);
    a->maxitask = b->maxitask;
    }
if (a->mintime > b->mintime) {
    a->mintime  = b->mintime;
    a->mintask  = b->mintask;
    a->mincount = b->mincount;
    a->mindata  = b->mindata;
    }
    
}

void PIiOutputStateCombined( fp, state, name, do_rate )
FILE    *fp;
SYstate *state;
char    *name;
int     do_rate;
{
#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
PIiStateSum in, work;

in.maxtime  = in.mintime  = state->time;
in.maxtask  = in.mintask  = PImytid;
in.maxcount = in.mincount = state->count;
in.maxdata  = in.mindata  = state->data;

in.maxline  = state->line;
in.maxitime = state->maxtime;
in.maxitask = PImytid;
strcpy( in.maxfile, state->file );

/* Get the min and max times */
PIcombine( &in, 1, &work, PSAllProcs, sizeof(PIiStateSum), MSG_OTHER, 
	   PIiStateSumOp );
if (PImytid == 0) {
    if (in.maxcount > 0) {
	fprintf( fp, 
"%8s min: %9d %10d %11.4le[%3d] \n\
         max: %9d %10d %11.4le[%3d]\n\
        [greatest time in %s (%d) on %d for %10.3le]\n", 
		name, in.mincount, in.mindata, in.mintime, in.mintask, 
		in.maxcount, in.maxdata, in.maxtime, in.maxtask, in.maxfile,
		in.maxline, in.maxitask, in.maxitime );
	}
    fflush( fp );
    }
PIgsync( PSAllProcs );
#endif
}	    

/* Clear all of the data in the summaries */
int PISummaryClear()
{
#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
_state_bsend.time  = 0.0;
_state_bsend.count = 0;
_state_bsend.data  = 0;
_state_bsend.maxtime = 0.0;
_state_nsend.time  = 0.0;
_state_nsend.count = 0;
_state_nsend.data  = 0;
_state_nsend.maxtime = 0.0;
_state_wsend.time  = 0.0;
_state_wsend.count = 0;
_state_wsend.data  = 0;
_state_wsend.maxtime = 0.0;

_state_brecv.time  = 0.0;
_state_brecv.count = 0;
_state_brecv.data  = 0;
_state_brecv.maxtime = 0.0;
_state_nrecv.time  = 0.0;
_state_nrecv.count = 0;
_state_nrecv.data  = 0;
_state_nrecv.maxtime = 0.0;
_state_wrecv.time  = 0.0;
_state_wrecv.count = 0;
_state_wrecv.data  = 0;
_state_wrecv.maxtime = 0.0;

_state_gop.time    = 0.0;
_state_gop.count   = 0;
_state_gop.data    = 0;
_state_gop.maxtime = 0.0;
_state_gsync.time  = 0.0;
_state_gsync.count = 0;
_state_gsync.data  = 0;
_state_gsync.maxtime = 0.0;
#endif
}

/*
   This routine generates a summary to the given output file for the 
   calling processor.
   So far, a reasonable way to use this might be:
   for (i=0; i<=PInumtids; i++)
       if (GTOKEN((ProcSet*)0,i)) {
           fprintf( stdout, "Summary for node %d:\n", PImytid );
	   PILogSummary( stdout );
	   fflush( stdout );
	   }

    An alternate approach could generate some kind of graphical summary,
    or several processors per line, like
 
    processor 0            1            2      ...
             time[#]<len>    ...

    Finally, communication rates and fits to models would be nice.  For
    example, for the simple latency + n*rate method, if we assume that
    the formula is accurate, and that there are m sends, then we have
    time = m * lat + rate * len .   We need one more equation.  For example,
    we could use
    sum(lat + rate*len[i]==t[i]) = time
    sum(t[i]/len[i]) = sum(lat/len[i] + rate) = m*rate + lat*sum(1/len[i])
    For this, we'd also have to save the sum of the inverses of the lengths, 
    and the sum of the times, scaled by the lengths.
 */
int PILogSummary( fp )
FILE *fp;
{
double r;
	
/* Do the most detailed preferentially */
#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
fprintf( fp, 
"Op:                      calls      bytes    total time Average rate\n" );
PIiOutputState( fp, &_state_bsend, "bsend",  1 );
PIiOutputState( fp, &_state_nsend, "nsend",  0 );
PIiOutputState( fp, &_state_wsend, "wsend",  1 );
PIiOutputState( fp, &_state_brecv, "brecv",  1 );
PIiOutputState( fp, &_state_nrecv, "nrecv",  0 );
PIiOutputState( fp, &_state_wrecv, "wrecv",  1 );
PIiOutputState( fp, &_state_gop,   "Global", 0 );
PIiOutputState( fp, &_state_gsync, "Gsync",  0 );

#elif defined(LOGCOMMCOUNTS) && !defined(LOGCOMMDISABLE)
fprintf( fp, "Op:     calls    bytes\n" );
fprintf( fp, "Send:   %10d %10d %13.4le\n", _logsend, _logsendn );
fprintf( fp, "Recv:   %10d %10d %13.4le\n", _logrecv, _logrecvn );

#endif

/* A more advanced estimate of the state and times could be done as
   follows.  Save not only the length of the message n_i and the time t_i,
   but also n_i^2 and n_i t_i.
   Define

   N0 = \sum 1
   N1 = \sum n_i
   N2 = \sum n_i^2
   T0 = \sum t_i
   T1 = \sum n_i t_i

   Then, if we model communication by s + rn, we have the equations

   T0 = \sum(s + r n_i) = s N0 + r N1
   T1 = \sum(n_i(s+r n_i)) = s N1 + r N2

   and the resulting estimate for s

       N2 T0 - N1 T1
   s = -------------
       N0 N2 - N1 N1

   and for r

       N0 T1 - N1 T0
   r = -------------
       N0 N2 - N1 N1

   Note that if all the n_i's are the same, the system is singular.
   It can also be very poorly conditioned, though we can compute how
   badly from the singular values (just the eigenvalues, since the
   matrix is symmetric).  The eigenvalues are

   (1/2) ( N2 + N0 +/- sqrt( (N2-N0)^2 + 4N1^2 ) )
   
   The ratio of these is the condition number of the matrix.  This
   value can be computed (carefully so as to avoid overflow); if it is
   small enough, the solution of the system is well-behaved.  Whether
   the results are meaningful is another matter...
 */
}

int PILogCombinedSummary( fp )
FILE *fp;
{
#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
if (PImytid == 0) 
    fprintf( fp, "Op:         calls      bytes        time task\n" );
PIiOutputStateCombined( fp, &_state_bsend, "bsend",  1 );
PIiOutputStateCombined( fp, &_state_nsend, "nsend",  0 );
PIiOutputStateCombined( fp, &_state_wsend, "wsend",  1 );
PIiOutputStateCombined( fp, &_state_brecv, "brecv",  1 );
PIiOutputStateCombined( fp, &_state_nrecv, "nrecv",  0 );
PIiOutputStateCombined( fp, &_state_wrecv, "wrecv",  1 );
PIiOutputStateCombined( fp, &_state_gop,   "Global", 0 );
PIiOutputStateCombined( fp, &_state_gsync, "Gsync",  0 );
#endif
}

/* 
   There are a number of places where we need to temporarily turn off 
   logging, then restore logging to its previous state.  This means that
   we need to remember the previous state and restore it (mostly, this
   happens around the global operations, but it would be nice if this
   worked correctly always
 */
#if !defined(LOGCOMMDISABLE)
#define MAX_STACK 25
static int logstatus[MAX_STACK+1];
static int logsp = 0;
void LOGpush()
{
if (logsp < MAX_STACK)
    logstatus[logsp++] = LOGSTATUS;
}
void LOGpop()
{
if (logsp > 0) logsp--;
LOGSTATUS = logstatus[logsp];
}
#else
/* Stubs in case these are used in their raw form (as they are in initp4) */
void LOGpush()
{}
void LOGpop()
{}
#endif 

#if !defined(LOGCOMMDISABLE)
/* This routine takes an event number and returns the name of the related op */
char *_PIEventName( n )
int n;
{
switch (n) {
    case LOG_GSUMS:
    case LOG_GSUME: return "GSUM";
    case LOG_GMAXS:
    case LOG_GMAXE: return "GMAX";
    case LOG_GMINS:
    case LOG_GMINE: return "GMIN";
    case LOG_GCOLS:
    case LOG_GCOLE: return "GCOL";
    case LOG_GSCATS:
    case LOG_GSCATE: return "GSCATTER";
    case LOG_GSYNCS:
    case LOG_GSYNCE: return "GSYNC";
    case LOG_GTOKS:
    case LOG_GTOKE: return "GTOKEN";
    }
return "Unknown";
}
#else
char *_PIEventName( n )
int n;
{
return "Unavailable event name";
}
#endif

/*@
   PISetTracefile - set the name of the file for tracing output

   Input Parameter:
.  name - name of trace file.  It may contain a "%d", in which case
          the %d is replaced with the value of PImytid.
@*/
void PISetTracefile( name )
char *name;
{
char filename[1024];

if (strchr( name, '%' )) {
    sprintf( filename, name, PImytid );
    _tracefile = fopen( filename, "w" );
    }
else
    _tracefile = fopen( name, "w" );

if (!_tracefile)
    _tracefile = stdout;
}

/*
   In order to simplify the collection and display of data on 
   communications, I'm introducing a few functions that 
   will be called instead of the current macros.

   These are

   PIiBSendStart( from, to, len, type, file, line )
   PIiBSendEnd(   from, to, len, type, file, line )
   PIiNSendStart( from, to, len, type, file, line )
   PIiNSendEnd(   from, to, len, type, file, line )
   PIiWSendStart( from, to, len, type, file, line )
   PIiWSendEnd(   from, to, len, type, file, line )

   PIiBRecvStart( to, type, stype, file, line )
   PIiBRecvEnd( from, to, len, type, stype, file, line )
   PIiNRecvStart( to, type, stype, file, line )
   PIiNRecvEnd(   to, type, stype, file, line )
   PIiWRecvStart( to, type, stype, file, line )
   PIiWRecvEnd( from, to, len, type, stype, file, line )

   These routine will contain all of the logging, state management, 
   event tracing, and so on.  

   The production versions will continue to have none of these.

   The usual use is
   PIiBSendStart(...) bsend() PIiBSendEnd(...)
   PIiBRecvStart(...) brecv() PIiBRecvEnd(...)
   ...

   Note that in the blocking receive case we want to capture the "idle" time.
   We do this by providing the option for a blocking probe on the
   type.  Each implementation needs to provide the appropriate macro for
   this (LOGRECVIDLEMACRO).

   Disadvantages:
   If these routines are used, you can't control the amount of logging
   at compile time.  

   The "file" and "line" args should be __FILE__ and __LINE__ .
   These allow error messages to be directed to the point where the
   routines were called.

   Finally, since we are calling routines here, we'll always provide
   all of the services (the additional overhead is slight)
 */

/* Redefine SETERRC to use "line,file" for the line and file. */
#undef SETERRC
#define SETERRC(a,s)      {__terrno.err = a;DEBUGTBCF(line,file,s);}

/* Macros to simplify the coding of the optional actions.  All of the
   form IAL...() */
#define IALEVENT(event) \
if (!(OnlyAtomic && inAtomicOp))    LOGEVENT(event);
#if defined(LOGCOMMEVENTS) && !defined(LOGCOMMDISABLE)
#define IALEVENT4(event,len,type,from) \
if (!(OnlyAtomic && inAtomicOp)) BLOG_LOG4(event,len,type,from); 
#else
#define IALEVENT4(event,len,type,from)
#endif

#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
#define IALSTATEPUSH(state,n) \
    if(!(OnlyAtomic && inAtomicOp) && (cloglevel & CLOG_SUMMARY)) \
    SYStatePush(state,n);
#define IALSTATEPOP(state,n,file,line) \
    if(!(OnlyAtomic && inAtomicOp) && (cloglevel & CLOG_SUMMARY)) \
    SYStatePop(state,n,file,line);
#else
#define IALSTATEPUSH(state,n)
#define IALSTATEPOP(state,n,file,line)
#endif

#define IALTRACE \
    (cloglevel & (CLOG_LASTOP|CLOG_TRACE)) && !((OnlyAtomic && inAtomicOp))
#define IALTRACEWRITE \
    if (cloglevel & CLOG_TRACE) { \
        fputs( buf, _tracefile ); \
        fflush(_tracefile);}

/* vfp... is basically a hook for arbitrary additional processing.
   The code is intended to be used to for a x-graphics virtual-front-panel,
   but the default is null to allow applications to run without X.
   
   The routine PIVFPInit() should be called to set up such a virtual front
   panel.
 */
static void (*vfpdraw)() = 0;
static void *vfpctx    = 0;

void PIiBSendStart( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
if (type < mintag || type > maxtag) { 
    sprintf( buf, "Invalid tag (%d) in %s", type, "bsend" );
    SETERRC(1,buf);
    }
/* Test for too large... */
if (len < 0) {
    sprintf( buf, "Invalid (negative) length (%d) in %s", len, "bsend" );
    SETERRC(1,buf);
    }
if (to < 0 || to >= PInumtids) {
    sprintf( buf, "Invalid destination (%d) in %s", to, "bsend" );
    SETERRC(1,buf);
    }

#if defined(LOGCOMMCOUNTS) && !defined(LOGCOMMDISABLE)
_logsend++ ;
_logsendn += len;
#endif

if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d(%d) To %d> \t[%s:%d]\n", 
	    PImytid, "bsend", type, len, to, file, line );
    IALTRACEWRITE;
    }
IALSTATEPUSH(&_state_bsend,len);

IALEVENT4(LOG_BSENDS,len,type,to)
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_BSENDS );
}

void PIiBSendEnd( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
IALEVENT(LOG_BSENDE);
IALSTATEPOP(&_state_bsend,0,file,line);
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_BSENDE );
}

void PIiNSendStart( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
if (type < mintag || type > maxtag) { 
    sprintf( buf, "Invalid tag (%d) in %s", type, "nsend" );
    SETERRC(1,buf);
    }
/* Test for too large... */
if (len < 0) {
    sprintf( buf, "Invalid (negative) length (%d) in %s", len, "nsend" );
    SETERRC(1,buf);
    }
if (to < 0 || to >= PInumtids) {
    sprintf( buf, "Invalid destination (%d) in %s", to, "nsend" );
    SETERRC(1,buf);
    }

#if defined(LOGCOMMCOUNTS) && !defined(LOGCOMMDISABLE)
_logsend++ ;
_logsendn += len;
#endif

if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d(%d) To %d> \t[%s:%d]\n", 
	    PImytid, "nsend", type, len, to, file, line );
    IALTRACEWRITE;
    }
IALSTATEPUSH(&_state_nsend,len);

IALEVENT4(LOG_NSENDS,len,type,to)
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_NSENDS );
}

void PIiNSendEnd( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
IALEVENT(LOG_NSENDE);
IALSTATEPOP(&_state_nsend,0,file,line);
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_NSENDE );
}

void PIiWSendStart( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
if (type < mintag || type > maxtag) { 
    sprintf( buf, "Invalid tag (%d) in %s", type, "wsend" );
    SETERRC(1,buf);
    }
/* Test for too large... */
if (len < 0) {
    sprintf( buf, "Invalid (negative) length (%d) in %s", len, "wsend" );
    SETERRC(1,buf);
    }
if (to < 0 || to >= PInumtids) {
    sprintf( buf, "Invalid destination (%d) in %s", to, "wsend" );
    SETERRC(1,buf);
    }

if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d(%d) To %d> \t[%s:%d]\n", 
	    PImytid, "wsend", type, len, to, file, line );
    IALTRACEWRITE;
    }
IALSTATEPUSH(&_state_wsend,len);

IALEVENT4(LOG_WSENDS,len,type,to)
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_WSENDS );
}

void PIiWSendEnd( to, len, type, file, line )
int     to, len, type, line;
char    *file;
{
IALEVENT(LOG_WSENDE);
IALSTATEPOP(&_state_wsend,0,file,line);
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, to, len, LOG_WSENDE );
}


/* This runs upto the actual receive.  It includes code to collect idle time */
void PIiBRecvStart( maxlen, type, file, line )
int  maxlen, type, line;
char *file;
{
if (type < mintag || type > maxtag) {
    sprintf( buf, "Invalid tag (%d) in brecv", type );
    SETERRC(1,buf);
    }
if (maxlen < 0 && maxlen != -1) {
    /* We use -1 for receive of unknown size */
    sprintf( buf, "Invalid (negative) max length (%d) in brecv", maxlen );
    SETERRC(1,buf);
    }

IALSTATEPUSH(&_state_brecv,0);
if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d> \t[%s:%d]\n", 
	     PImytid, "brecv(start)", type, file, line );
    IALTRACEWRITE;
    }
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, -1, -1, LOG_BRECVS );

#ifdef LOGRECVIDLE
if (cloglevel & CLOG_IDLE) {
    /* idle wait event start */
    IALEVENT(LOG_IRECVS);
#ifdef LOGRECVIDLEMACRO
    LOGRECVIDLEMACRO(type);
#else  
    SYNCPROBE(type);
#endif    
    /* idle wait event end */
    IALEVENT(LOG_IRECVE);
    }
#endif
IALEVENT(LOG_BRECVS);
}

void PIiBRecvEnd( from, len, type, file, line )
int  from, len, type, line;
char *file;
{
IALEVENT4(LOG_BRECVE,len,type,from);
IALSTATEPOP(&_state_brecv,len,file,line);
#if defined(LOGCOMMCOUNTS) && !defined(LOGCOMMDISABLE)
_logrecv++; 
_logrecvn += len;
#endif
if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d(%d) From %d> \t[%s:%d]\n", 
	    PImytid, "brecv(end)", type, len, from, file, line );
    IALTRACEWRITE;
    }

if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, from, PImytid, len, LOG_BRECVE );
}

void PIiNRecvStart( maxlen, type, file, line )
int  maxlen, type, line;
char *file;
{
if (type < mintag || type > maxtag) {
    sprintf( buf, "Invalid tag (%d) in nrecv", type );
    SETERRC(1,buf);
    }
if (maxlen < 0) {
    sprintf( buf, "Invalid (negative) max length (%d) in nrecv", maxlen );
    SETERRC(1,buf);
    }

IALSTATEPUSH(&_state_nrecv,0);
if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d> \t[%s:%d]\n", 
	     PImytid, "nrecv", type, file, line );
    IALTRACEWRITE;
    }
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, -1, -1, LOG_NRECVS );
IALEVENT(LOG_NRECVS);
}

void PIiNRecvEnd( type, file, line )
int  type, line;
char *file;
{
IALSTATEPOP(&_state_nrecv,0,file,line);
IALEVENT(LOG_NRECVE);
}

void PIiWRecvStart( type, file, line )
int  type, line;
char *file;
{
IALSTATEPUSH(&_state_wrecv,0);
if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d> \t[%s:%d]\n", 
	     PImytid, "wrecv(start)", type, file, line );
    IALTRACEWRITE;
    }
if (vfpdraw)
    (*vfpdraw)( vfpctx, PImytid, PImytid, -1, -1, LOG_WRECVS );
IALEVENT(LOG_WRECVS);
}

void PIiWRecvEnd( from, len, type, file, line )
int  from, len, type, line;
char *file;
{
IALSTATEPOP(&_state_wrecv,len,file,line);
if(IALTRACE) {
    sprintf( buf, "[%d] %s <Tag %d(%d) From %d> \t[%s:%d]\n", 
	    PImytid, "wrecv(end)", type, len, from, file, line );
    IALTRACEWRITE;
    }
IALEVENT(LOG_WRECVE);
}

void PISetVFP( vctx, vdraw )
void *vctx, (*vdraw)();
{
vfpctx  = vctx;
vfpdraw = vdraw;
}

/* This routine should be called by an error handler to print the
   last trace message that would have been generated by -trace
   I'm not sure yet how to do that (though establising a chain of
   error handlers isn't a bad idea) */
void PIPrintLastTraceMessage()
{
if(cloglevel & CLOG_LASTOP) {
    buf[255] = 0;  /* just in case */
    fputs( buf, _tracefile );
    }
}

/*
    Here is code to keep track of the location of the maximum time
    used in a communication operation.

    For this we need:
    the time for this last op, 
    the current MAXIMUM
    
    And a place to put all of this data
 */
