/*
   This is a test program for the fd write and read routines

   This tries to maximize performance
 */

#include "tools.h"
#include "system/system.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#include <math.h>

int main( argc, argv )
int  argc;
char **argv;
{
int worker();
PICall( worker, argc, argv );
}

int worker( argc, argv )
int  argc;
char **argv;
{
PIFILE      *fp;
int         n, np, nd;
double      *v, *v2;
double      t1, t2;
int         fsize, bytes, nsize, myid, j, nb;
int         csize = 0;
char        fname[1000];
int         err;
int         pmode;

strcpy( fname, "foo" );
/* get the global size */
if (SYArgHasName( &argc, argv, 1, "-help" )) {
    fprintf( stderr, 
"%s -n n -csize n -fsize n -f name -pio\n\
      -n n      Size of each block to write/read\n\
      -csize n  Size of read/write cache\n\
      -fsize    Size of file in doubles\n\
      -f name   Name of file\n\
      -pio      Use parallel IO (PIO_AS_PARALLEL)\n", argv[0] );
    return 0;
    }

SYArgGetString( &argc, argv, 1, "-f", fname, 1000 );
fsize = 1024;
SYArgGetInt( &argc, argv, 1, "-fsize", &fsize );
n    = fsize / PInumtids;
SYArgGetInt( &argc, argv, 1, "-n", &n );
SYArgGetInt( &argc, argv, 1, "-csize", &csize );
pmode = PIO_AS_SEQUENTIAL;
if (SYArgHasName( &argc, argv, 1, "-pio" )) pmode = PIO_AS_PARALLEL;

if (PImytid == 0) 
    printf( "# np  cache  blocksize   filesize Wrate(MB/s) Rrate(MB/s)\n" );
while (n <= fsize / PInumtids ) {
    fp = PIFopen( fname, PIAllProcs, O_WRONLY | O_CREAT, pmode, -1 );
    /* Switch to cached */
    if (csize > 0) 
	PIFCacheEnable( fp, csize, csize );
    
    /*
      This is the test.  Each processor allocates fsize/np data.
      This data is read/written in blocks of size n
      */
    np    = PInumtids;
    nsize = fsize / np;
    v     = (double *)MALLOC( nsize * sizeof(double) );  CHKPTR(v);
    myid  = PImytid;
    for (j=0; j<nsize; j++) 
	v[j] = myid * nsize + j;
    nb = nsize / n;
    PIgsync( PIAllProcs );
    t1 = SYGetElapsedTime();
    for (j=0; j<nb; j++) 
	PIFwrite( fp, v + j * n, n * sizeof(double), MSG_DBL );
    PIFclose( fp );
    PIgsync( PIAllProcs );
    t2 = SYGetElapsedTime() - t1;
    if (PImytid == 0) {
	bytes = fsize * sizeof(double);
	printf( "%4d %6d %10d %10d %11.3f ", 
	        np, csize, n*sizeof(double), bytes, bytes/(1.0e6*t2) );
	fflush( stdout );
	}
    
    /* Now, try to READ the data back; check that it is correct */
    fp = PIFopen( fname, PIAllProcs, O_RDONLY, pmode, -1 );
    if (!fp) {
	fprintf( stderr, "Can not open file %s for input\n", fname );
	return 0;
	}
    if (csize > 0) 
	PIFCacheEnable( fp, csize, csize );
    v2 = (double *)MALLOC( nsize * sizeof(double) );  CHKPTR(v2);

    PIgsync( PIAllProcs );
    t1 = SYGetElapsedTime();
    for (j=0; j<nb; j++) 
	PIFread( fp, v2 + j * n, n * sizeof(double), MSG_DBL );
    PIFclose( fp );
    PIgsync( PIAllProcs );
    t2 = SYGetElapsedTime() - t1;
    if (PImytid == 0) {
	bytes = fsize * sizeof(double);
	printf( "%11.3f\n", bytes/(1.0e6*t2) );
	fflush( stdout );
	}
    
    /* Check that the values are correct */
    err = 0;
    for (j=0; j<nsize; j++) {
	if (v[j] != v2[j]) {
	    fprintf( stderr, 
		   "[%d]Error in V values in file in location %d (%e != %e)\n",
		    PImytid, j, v[j], v2[j] );
	    err++;
	    if (err > 10) {
		fprintf( stderr, "Error number exceeded; exiting\n" );
		return err;
		}
	    }
	}


    FREE( v );
    FREE( v2 );
    n  = n * 2;
    }
return 0;
}
