/* 
   Communications activity is logged by default.  The number of sends and
   receives, and the amount of data, is all logged.  Turn this off by 
   defining LOGCOMMDISABLE

   This has gotten out-of-hand.  We want to define the following
   macros:
   LOGBSENDSTART(len,type,to)
   LOGBSENDEND(type)
   LOGNSENDSTART(len,type,to)
   LOGNSENDEND(type)
   LOGWSENDSTART(len,type,to)
   LOGWSENDEND(type)

   
   LOGBRECVSTART(type)
   LOGBRECVEND(type,len,from)
   LOGNRECVSTART(type)
   LOGNRECVEND(type)
   LOGWRECVSTART(type)     
   LOGWRECVEND(type,len,from)

   (These match the primative Chameleon operations.) Each of these in turn 
   contains 3 separate calls: 
   tracing
   state-counting
   event-logging

   Note that for receive, we may not have the len and from data easily
   available.  This is the rule:
   if eventlogging is enabled, we get the values and store them; 
   the RECVLEN etc routines just access these values rather than get the
   values from the implementation layer

 */

/* Used to control run-time logging */
#ifndef __LOGGING
#define __LOGGING

extern int cloglevel;
extern int OnlyAtomic, inAtomicOp;

#if defined(LOGCOMMALL)
#define LOGCOMMTRACE
#define LOGCOMMCOUNTS
#define LOGCOMMEVENTS
#define LOGCOMMSUMMARY
#endif

/* Tracing generates output for each send/recv */
#if defined(LOGCOMMTRACE) && !defined(LOGCOMMDISABLE)
# include <stdio.h>
extern FILE *_tracefile;

#define LPRINTS(a,type,len,to) {if(cloglevel&0x2){\
	   fprintf( _tracefile, "[%d] %s <Type %d(%d) To %d> [%s:%d]\n", \
		  MYPROCID, a, type, len, to, __FILE__, __LINE__ );\
		fflush(_tracefile);}}
#define LPRINTR(a,type) {if(cloglevel&0x2){\
           fprintf( _tracefile, "[%d] %s <Type %d> [%s:%d]\n", \
		  MYPROCID, a, type, __FILE__, __LINE__ );fflush(_tracefile);}}
extern char *_PIEventName();
#define LPRINTG(a,nam) {if(cloglevel&0x2){\
	 fprintf( _tracefile, "[%d] %s <%s> [%s:%d]\n", \
 	 MYPROCID, a, _PIEventName(nam), __FILE__, __LINE__ );\
	     fflush(_tracefile);}}
#else
#define LPRINTS(a,b,c,d)
#define LPRINTR(a,b)
#define LPRINTG(a,n)
#endif

/* Count the number of sends and receives, and the sizes */
#if defined(LOGCOMMCOUNTS) && !defined(LOGCOMMDISABLE)
extern int _logsend, _logsendn, _logrecv, _logrecvn;
#define LOGSENDCNT(n,type,to){_logsend++;_logsendn+=n;\
				  LPRINTS("send",type,n,to);}
#define LOGRECVCNT(n,type) {_logrecv++; _logrecvn+=n;LPRINTR("recv",type);}
#else
#define LOGSENDCNT(n,type,to) LPRINTS("send",type,n,to)
#define LOGRECVCNT(n,type)    LPRINTR("recv",type)
#endif

#include "system/state.h"
#if defined(LOGCOMMSUMMARY) && !defined(LOGCOMMDISABLE)
extern SYstate _state_bsend, _state_brecv, _state_nsend, _state_nrecv, 
               _state_wsend, _state_wrecv,  
               _state_gsync, _state_gop;
#define LOGSTATESTART(state,n) if(cloglevel&0x4)SYStatePush(state,n)
#define LOGSTATEEND(state,n)   if(cloglevel&0x4)\
    SYStatePop(state,n,__FILE__,__LINE__)
#else
#define LOGSTATESTART(state,n)
#define LOGSTATEEND(state,n)
#endif

/* Event logging */
/* The choices of event numbers.  I should allow the user to control 
   these... */
#define LOG_BSENDS 103
#define LOG_BSENDE 104
#define LOG_NSENDS 105
#define LOG_NSENDE 106
#define LOG_WSENDS 107
#define LOG_WSENDE 108
#define LOG_BRECVS 109
#define LOG_BRECVE 110
#define LOG_NRECVS 111
#define LOG_NRECVE 112
#define LOG_WRECVS 113
#define LOG_WRECVE 114
#define LOG_IRECVS 115
#define LOG_IRECVE 116

#define LOG_GSUMS 120
#define LOG_GSUME 121
#define LOG_GMAXS 122
#define LOG_GMAXE 123
#define LOG_GMINS 124
#define LOG_GMINE 125
#define LOG_GCOLS 126
#define LOG_GCOLE 127
#define LOG_GSCATS 128
#define LOG_GSCATE 129
#define LOG_GSYNCS 130
#define LOG_GSYNCE 131
#define LOG_GTOKS  132
#define LOG_GTOKE  133

#if defined(LOGCOMMEVENTS) && !defined(LOGCOMMDISABLE)
/* Routines to push and pop the event logging status (in ialog) */
extern void LOGpush(), LOGpop();

/* We used to support alog and blog.  We've removed the support for alog since
   blog is (mostly) a superset, and it has been a long time since the alog
   support has been tested (and alog is no longer included with the 
   distribution). */

/* We are setup to use 2 different event logging systems; BLOG is a 
   simplified ALOG */
#include "blog/blog.h"
#define LOGEVENT(a)    BLOG_LOG(a)
#define LOGOPSTART(n) {LPRINTG("Starting",n);LOGSTATESTART(&_state_gop,0);\
			   LOGEVENT(n);LOGpush();BLOG_DISABLE;}
#define LOGOPEND(n)   {LOGpop();LOGEVENT(n);LOGSTATEEND(&_state_gop,0);\
			   LPRINTG("Ending",n);}
#define LOGSYNCSTART(n) {LPRINTG("Starting",n);LOGSTATESTART(&_state_gsync,0);\
			   LOGEVENT(n);LOGpush();BLOG_DISABLE;}
#define LOGSYNCEND(n)   {LOGpop();LOGEVENT(n);LOGSTATEEND(&_state_gsync,0);\
			   LPRINTG("Ending",n);}
#define LOGDEFINE(a,b) BLOG_DEFINE(a,b)
#define LOGDISABLE     BLOG_DISABLE
#define LOGENABLE      BLOG_ENABLE
#define LOGSTATE(a,b,c,s) BLOG_DEFINE_STATE(a,b,c,s)
#define LOGSTATUS      xx_BLOG_status

#else
#define LOGEVENT(a)
#define LOGOPSTART(n)      {LPRINTG("Starting",n);\
				LOGSTATESTART(&_state_gop,0);}
#define LOGOPEND(n)        {LOGSTATEEND(&_state_gop,0);LPRINTG("Ending",n);}
#define LOGSYNCSTART(n) {LPRINTG("Starting",n);LOGSTATESTART(&_state_gsync,0);}
#define LOGSYNCEND(n)   {LOGSTATEEND(&_state_gsync,0);LPRINTG("Ending",n);}
#define LOGDISABLE  
#define LOGENABLE   
#define LOGDEFINE(a,b)
#define LOGSTATE(a,b,c,s)
#endif

/* Here are the events that will be used to synchronize clocks */
#define EVENT_SYNC        -101
#define EVENT_PAIR_A1     -102
#define EVENT_PAIR_A2     -103
#define EVENT_PAIR_B1     -104

/*
   Now that we have defined all of the individual logging operations, 
   we define the ones to use in the communications operations.
   Note that since the operations that need to be performed for full
   logging are fairly complicated, we put them inside of routines that 
   are defined in ialog.c .  These routines handle logging, events, states, 
   and the virtual front panel, if enabled.  Note that if logging is diabled,
   the routines are not even called, ensuring that there is no overhead in
   the production version.
 */
#if defined(LOGCOMMDISABLE)
#define LOGBSENDSTART(len,type,to) 
#define LOGBSENDEND(type) 
#define LOGNSENDSTART(len,type,to) 
#define LOGNSENDEND(type) 
#define LOGWSENDSTART(len,type,to) 
#define LOGWSENDEND(type) 

#define LOGBRECVSTART(maxlen,type) 
#define LOGBRECVEND(type,len,from) 
#define LOGNRECVSTART(maxlen,type) 
#define LOGNRECVEND(type) 
#define LOGWRECVSTART(type) 
#define LOGWRECVEND(type,len,from) 

#else
#define LOGBSENDSTART(len,type,to) \
    PIiBSendStart( to, len, type, __FILE__, __LINE__ )
#define LOGBSENDEND(type)   \
    PIiBSendEnd( 0, 0, type, __FILE__, __LINE__ )
#define LOGNSENDSTART(len,type,to) \
    PIiNSendStart( to, len, type, __FILE__, __LINE__ )
#define LOGNSENDEND(type)   \
    PIiNSendEnd( 0, 0, type, __FILE__, __LINE__ )
#define LOGWSENDSTART(len,type,to) \
    PIiWSendStart( to, len, type, __FILE__, __LINE__ )
#define LOGWSENDEND(type)   \
    PIiWSendEnd( 0, 0, type, __FILE__, __LINE__ )

#define LOGBRECVSTART(maxlen,type) \
    PIiBRecvStart( maxlen, type, __FILE__, __LINE__ )
#define LOGBRECVEND(type,len,from)   \
    PIiBRecvEnd( from, len, type, __FILE__, __LINE__ )
#define LOGNRECVSTART(maxlen,type) \
    PIiNRecvStart( maxlen, type, __FILE__, __LINE__ )
#define LOGNRECVEND(type)   \
    PIiNRecvEnd( type, __FILE__, __LINE__ )
#define LOGWRECVSTART(type) \
    PIiWRecvStart( type, __FILE__, __LINE__ )
#define LOGWRECVEND(type,len,from)   \
    PIiWRecvEnd( from, len, type, __FILE__, __LINE__ )

#endif

/* Logging control for atomic operations.  Users should NOT use these unless
   they are very, very sure about what they are doing */
#if defined(LOGCOMMALL) || defined(LOGCOMMTRACE) || defined(LOGCOMMEVENTS)
#define LOGPUSHATOMIC inAtomicOp++
#define LOGPOPATOMIC  inAtomicOp--
#else
#define LOGPUSHATOMIC
#define LOGPOPATOMIC 
#endif

#endif /*__LOGGING*/
